package com.vaadin.incubator.dashlayout.client.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import com.google.gwt.dom.client.Document;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.ui.ComplexPanel;
import com.google.gwt.user.client.ui.Widget;
import com.vaadin.terminal.gwt.client.ApplicationConnection;
import com.vaadin.terminal.gwt.client.BrowserInfo;
import com.vaadin.terminal.gwt.client.Container;
import com.vaadin.terminal.gwt.client.Paintable;
import com.vaadin.terminal.gwt.client.RenderSpace;
import com.vaadin.terminal.gwt.client.UIDL;
import com.vaadin.terminal.gwt.client.ValueMap;
import com.vaadin.terminal.gwt.client.RenderInformation.FloatSize;
import com.vaadin.terminal.gwt.client.RenderInformation.Size;
import com.vaadin.terminal.gwt.client.ui.AlignmentInfo;
import com.vaadin.terminal.gwt.client.ui.VMarginInfo;

/**
 * This layout tries to fulfill two different tasks:
 * <p>
 * 1. Provide theme designers a more flexible and efficient layout in terms of
 * margin, border and padding.
 * <p>
 * 2. Implement the most basic layout with lightning fast performance: an
 * undefined sized vertical ordered layout without expand ratio functionalities.
 * All other features remain: margins, spacings and sizing. Sizing will simply
 * clip the overflowing components.
 * <p>
 * These tasks require different kind of DOM setups (same structure, different
 * properties), hence two separate operation modes. Changing from one to the
 * other causes a total reflow of the layout.
 * <p>
 * The first implementation (MODE_FULL) is designed NOT to support adding
 * components efficiently. When a new component is added (after the initial
 * paint), all component locations and sizes are recalculated from scratch
 * (meaning a complete reflow).
 * <p>
 * The second implementation (MODE_STATIC pure stacking layout) will be able to
 * quickly insert new widgets anywhere in the layout.
 * <p>
 * And BTW, this layout is only vertical. Horizontal version is a different
 * class.
 */
public class VDashLayout extends ComplexPanel implements Container {

    public static final String CLASSNAME = "v-weelayout";

    private int mode = -1;
    public static final int MODE_FULL = 0;
    public static final int MODE_STATIC = 1;

    /**
     * A logical mapping from widget to its virtual layout cell.
     */
    private Map<Widget, ChildCell> widgetToCell = new HashMap<Widget, ChildCell>();

    /* Current spacing size */
    // private int spacing = 0;
    /**
     * All array values are in the following order: top, right, bottom, left.
     */

    /* Current margin values */
    private int[] margin = { 0, 0, 0, 0 };

    /* Current padding values */
    private int[] padding = { 0, 0, 0, 0 };

    /* Current border sizes */
    private int[] border = { 0, 0, 0, 0 };

    /*
     * Current inner size (excluding margins, border and padding) in pixels
     */
    private int width = 0;
    private int height = 0;

    private boolean isRendering;

    // Is the layout width undefined, i.e. defined by the contained components
    private boolean dynamicWidth = false;

    // We need to maintain the widest component for reference if the layout has
    // dynamic width. This value is be used to recognize situations when the
    // dynamic layout size is changed.
    // private int widestChild;

    protected boolean sizeHasChangedDuringRendering = false;

    private ApplicationConnection client;

    private String lastStyleName;

    private VMarginInfo marginInfo;

    private float compoundRatio = -1;

    private int consumedHeight = 0;

    public VDashLayout() {
        super();
        setElement(Document.get().createDivElement());
        setStylePrimaryName(CLASSNAME);
        getElement().getStyle().setProperty("float", "left");
        getElement().getStyle().setProperty("cssFloat", "left");
        getElement().getStyle().setProperty("styleFloat", "left");
        getElement().getStyle().setProperty("overflow", "hidden");
        getElement().getStyle().setProperty("display", "inline");
        getElement().getStyle().setProperty("zoom", "1");
    }

    private void setMode(int mode) {
        if (this.mode == -1) {
            this.mode = mode;
            return;
        }
        this.mode = mode;
        forceUpdateChildCells();
    }

    public void updateFromUIDL(UIDL uidl, ApplicationConnection client) {
        this.client = client;

        // Only non-cached UIDL:s can introduce changes
        if (uidl.getBooleanAttribute("cached")) {
            return;
        }

        // Margins need to be set before
        int bitMask = uidl.getIntAttribute("margins");
        if (marginInfo == null || marginInfo.getBitMask() != bitMask) {
            marginInfo = new VMarginInfo(bitMask);
            updateMargins();
        }

        // interpret as requestRepaint requested. Could be just an update on
        // margins or expandratios, but we need to update the margin, border and
        // padding values in this case
        if ((uidl.hasAttribute("mode") && uidl.getIntAttribute("mode") == mode)) {
            measureMarginsAndSpacing();
        }

        if (client.updateComponent(this, uidl, true)) {
            return;
        }

        isRendering = true;
        // Only non-cached, visible UIDL:s can introduce changes
        if (uidl.getBooleanAttribute("cached")
                || uidl.getBooleanAttribute("invisible")) {
            isRendering = false;
            return;
        }

        setMode(uidl.getIntAttribute("mode"));

        updateDynamicWidthInfo(uidl);

        // If the layout has dynamic width, we need to update all relative sized
        // components after all components have been rendered. Alignments will
        // also need to be updated.
        final ArrayList<ChildCell> updateAfter = new ArrayList<ChildCell>();

        // Iterate through Paintables in UIDL, add new ones and remove any
        // old ones.
        final int uidlCount = uidl.getChildCount();
        final int layoutCount = getChildren().size();
        int uidlPos = 0;
        int layoutPos = 0;

        // Additional info that needs to be passed to components
        final ValueMap alignments = uidl.getMapAttribute("alignments");
        final ValueMap expandRatios = uidl.getMapAttribute("expandRatios");

        if (expandRatios.getKeySet().size() > 0) {
            compoundRatio = 0;
        } else {
            compoundRatio = -1;
        }

        for (; (uidlPos < uidlCount || layoutPos < layoutCount); uidlPos++) {

            final UIDL childUIDL = (uidlPos < uidlCount) ? uidl
                    .getChildUIDL(uidlPos) : null;

            final Widget uidlWidget = childUIDL != null ? (Widget) client
                    .getPaintable(childUIDL) : null;

            final Widget layoutWidget = (layoutPos < layoutCount) ? getChildren()
                    .get(layoutPos)
                    : null;

            // layout position is past UIDL position: all remaining widgets
            // in layout are old ones, and need to be removed
            if (uidlPos >= uidlCount && layoutPos < layoutCount) {
                removeChildrenAfter(layoutPos);
                break;
            }

            // Old widget in old location, all OK
            if (uidlWidget == layoutWidget) {
                layoutPos++;
            }

            // Widget is either new or has changed place

            else if (getChildren().contains(uidlWidget)) {
                // An old component has changed place
                if (getChildren().indexOf(uidlWidget) != layoutPos) {
                    // Detach from old position child.
                    uidlWidget.removeFromParent();

                    // Logical attach.
                    getChildren().insert(uidlWidget, layoutPos);

                    getElement().insertBefore(uidlWidget.getElement(),
                            getElement().getChildNodes().getItem(layoutPos));

                    adopt(uidlWidget);
                    layoutPos++;
                }
            } else {
                // A completely new widget is inserted to this position
                ChildCell cell = new ChildCell(uidlWidget);
                widgetToCell.put(uidlWidget, cell);

                // Logical attach
                getChildren().insert(uidlWidget, layoutPos);

                // Avoid inserts (they are slower than appends)
                if (layoutPos < getChildren().size() - 1) {
                    getElement().insertBefore(uidlWidget.getElement(),
                            getElement().getChildNodes().getItem(layoutPos));
                } else {
                    getElement().appendChild(uidlWidget.getElement());
                }
                layoutPos++;
                // Adopt
                adopt(uidlWidget);
            }
            ChildCell cell = widgetToCell.get(uidlWidget);
            cell.updateSizeInfo(childUIDL);
            cell.updateSpace();

            ((Paintable) uidlWidget).updateFromUIDL(childUIDL, client);

            if (alignments.containsKey(childUIDL.getId())) {
                cell.setAlignment(alignments.getInt(childUIDL.getId()));
            }
            if (mode == MODE_FULL) {
                if (expandRatios.containsKey(childUIDL.getId())) {
                    final float ratio = expandRatios.getInt(childUIDL.getId());
                    compoundRatio += ratio;
                    cell.setExpandRatio(ratio);
                }
            }

            if (cell.updateAfterOthers()) {
                updateAfter.add(cell);
                if (!cell.isRelativeHeight()) {
                    cell.updateNaturalSize();
                }
            } else {
                cell.updateNaturalSize();
                cell.updateWidgetMargin();
                cell.reAlign();
            }

        } // All UIDL widgets painted

        // All remaining widgets are removed
        removeChildrenAfter(layoutPos);

        if (mode == MODE_FULL) {
            consumedHeight = calculateUsedHeight();
        }

        if (updateAfter.size() > 0) {
            int totalHeight = consumedHeight;
            for (int i = 0; i < updateAfter.size() - 1; i++) {
                ChildCell cell = updateAfter.get(i);
                cell.updateWidgetMargin();
                cell.updateSpace();
                if (cell.isRelativeWidth() || cell.isRelativeHeight()) {
                    client.handleComponentRelativeSize(cell.getWidget());
                    cell.updateNaturalSize();
                }
                cell.reAlign();
                if (cell.isRelativeHeight() || cell.expandRatio != -1) {
                    totalHeight += cell.getRenderSpace().getHeight();
                }
            }

            // Add rounding error pixel to last child
            ChildCell last = updateAfter.get(updateAfter.size() - 1);
            last.updateWidgetMargin();
            last.updateSpace();
            last.getRenderSpace().setHeight(height - totalHeight);
            if (last.isRelativeWidth() || last.isRelativeHeight()) {
                client.handleComponentRelativeSize(last.getWidget());
                last.updateNaturalSize();
            }
            last.reAlign();
        }

        if (dynamicWidth) {
            updateActualWidth();
        }

        isRendering = false;

    }

    private void updateActualWidth() {
        width = parsePixel(getStyleValue(getElement(), "width"));
    }

    private void updateMargins() {
        margin = collectMargin(getElement());
        if (marginInfo.hasTop()) {
            getElement().getStyle().setProperty("marginTop", "");
        } else {
            getElement().getStyle().setProperty("marginTop", "0");
            margin[0] = 0;
        }
        if (marginInfo.hasRight()) {
            getElement().getStyle().setProperty("marginRight", "");
        } else {
            getElement().getStyle().setProperty("marginRight", "0");
            margin[1] = 0;
        }
        if (marginInfo.hasBottom()) {
            getElement().getStyle().setProperty("marginBottom", "");
        } else {
            getElement().getStyle().setProperty("marginBottom", "0");
            margin[2] = 0;
        }
        if (marginInfo.hasLeft()) {
            getElement().getStyle().setProperty("marginLeft", "");
        } else {
            getElement().getStyle().setProperty("marginLeft", "0");
            margin[3] = 0;
        }
    }

    private void updateDynamicWidthInfo(UIDL uidl) {
        String w = uidl.hasAttribute("width") ? uidl
                .getStringAttribute("width") : "";

        if (w.equals("")) {
            dynamicWidth = true;
        } else {
            dynamicWidth = false;
        }
    }

    protected void removeChildrenAfter(int pos) {
        int toRemove = getChildren().size() - pos;
        while (toRemove-- > 0) {
            Widget child = getChildren().get(pos);
            widgetToCell.remove(child);
            remove(child);
            client.unregisterPaintable((Paintable) child);
        }
    }

    private void measureMarginsAndSpacing() {
        margin = collectMargin(getElement());
        padding = collectPadding(getElement());
        border = collectBorder(getElement());

        // TODO Measure spacing somehow
    }

    private int calculateUsedHeight() {
        int used = 0;
        for (Widget w : getChildren()) {
            ChildCell cell = widgetToCell.get(w);
            if (cell.expandRatio == -1 && !cell.isRelativeHeight()) {
                used += cell.getWidgetSize().getHeight();
            }
        }
        return used;
    }

    /**
     * Container interface methods
     */

    public RenderSpace getAllocatedSpace(Widget child) {
        ChildCell cell = widgetToCell.get(child);
        if (child instanceof VDashLayout) {
            // VFlexLayout handles it's margins internally
            return cell.getRenderSpace();
        }
        return cell.getSpaceSansMargins();
    }

    public boolean hasChildComponent(Widget component) {
        return getChildren().contains(component);
    }

    public void replaceChildComponent(Widget oldComponent, Widget newComponent) {
        if (!hasChildComponent(oldComponent)) {
            return;
        }
        getChildren().insert(newComponent, getChildren().indexOf(oldComponent));
        getChildren().remove(oldComponent);
        getElement().insertBefore(newComponent.getElement(),
                oldComponent.getElement());
        getElement().removeChild(oldComponent.getElement());
        orphan(oldComponent);
        adopt(newComponent);
        widgetToCell.remove(oldComponent);
        ChildCell cell = new ChildCell(newComponent);
        widgetToCell.put(newComponent, cell);
    }

    public boolean requestLayout(Set<Paintable> children) {
        for (Paintable child : children) {
            ChildCell cell = widgetToCell.get(child);
            cell.updateSpace();
            cell.updateNaturalSize();
            cell.reAlign();
        }
        if (dynamicWidth) {
            final int oldWidth = width;
            updateActualWidth();
            if (width != oldWidth) {
                for (Widget child : getChildren()) {
                    ChildCell childCell = widgetToCell.get(child);
                    childCell.updateSpace();
                    childCell.updateNaturalSize();
                    childCell.reAlign();
                }
                return false;
            }
        }
        return true;
    }

    public void updateCaption(Paintable component, UIDL uidl) {
        // TODO Use this same class to display component captions
        // TODO I need to make this class usable without the server-side
    }

    private void updateChildCells(boolean resetChildCells) {
        ArrayList<ChildCell> updateAfter = new ArrayList<ChildCell>();
        for (Widget child : getChildren()) {
            ChildCell cell = widgetToCell.get(child);
            if (resetChildCells) {
                cell.reset();
            }
            if (cell.updateAfterOthers()) {
                updateAfter.add(cell);
                if (!cell.isRelativeHeight()) {
                    cell.updateWidgetMargin();
                    cell.updateNaturalSize();
                }
            } else {
                cell.updateWidgetMargin();
                cell.updateSpace();
                client.handleComponentRelativeSize(child);
                cell.updateNaturalSize();
                cell.reAlign();
            }
        }
        if (mode == MODE_FULL) {
            consumedHeight = calculateUsedHeight();
        }
        if (updateAfter.size() == 0) {
            return;
        }
        int totalHeight = consumedHeight;
        for (int i = 0; i < updateAfter.size() - 1; i++) {
            ChildCell cell = updateAfter.get(i);
            cell.updateWidgetMargin();
            cell.updateSpace();
            if (cell.isRelativeWidth() || cell.isRelativeHeight()) {
                client.handleComponentRelativeSize(cell.getWidget());
                cell.updateNaturalSize();
            }
            cell.reAlign();
            if (cell.isRelativeHeight() || cell.expandRatio != -1) {
                totalHeight += cell.getRenderSpace().getHeight();
            }
        }

        // Add rounding error pixel to last child
        ChildCell last = updateAfter.get(updateAfter.size() - 1);
        last.updateWidgetMargin();
        last.updateSpace();
        last.getRenderSpace().setHeight(height - totalHeight);
        if (last.isRelativeWidth() || last.isRelativeHeight()) {
            client.handleComponentRelativeSize(last.getWidget());
            last.updateNaturalSize();
        }
        last.reAlign();
    }

    private void forceUpdateChildCells() {
        updateChildCells(true);
        updateActualWidth();
    }

    @Override
    public void setStyleName(String styleName) {
        super.setStyleName(styleName);

        if (isAttached() && !styleName.equals(lastStyleName)) {
            measureMarginsAndSpacing();
            updateChildCells(false);
            lastStyleName = styleName;
        }

    }

    /**
     * Only pixel values are accepted.
     */
    @Override
    public void setWidth(String w) {
        // Assume pixel values are always passed from ApplicationConnection

        int oldW = width;

        if (w != null && w != "") {
            int width = parsePixel(w) - margin[1] - margin[3] - border[1]
                    - border[3] - padding[1] - padding[3];
            if (width < 0) {
                width = 0;
            }
            super.setWidth(width + "px");
        } else {
            width = -1;
            super.setWidth("");
            forceUpdateChildCells();
        }
        updateActualWidth();

        if (width != oldW) {
            if (isRendering) {
                sizeHasChangedDuringRendering = true;
            }
            updateChildCells(false);
            client.runDescendentsLayout(this);
        }
    }

    /**
     * Only pixel values are accepted.
     */
    @Override
    public void setHeight(String h) {
        // Assume pixel values are always passed from ApplicationConnection

        int oldH = height;

        if (h != null && h != "") {
            int height = parsePixel(h) - margin[0] - margin[2] - border[0]
                    - border[2] - padding[0] - padding[2];
            if (height < 0) {
                height = 0;
            }
            this.height = height;
            super.setHeight(height + "px");
        } else {
            height = -1;
            super.setHeight("");
        }

        if (height != oldH) {
            if (isRendering) {
                sizeHasChangedDuringRendering = true;
            }
            if (mode == MODE_FULL) {
                updateChildCells(false);
            }
            client.runDescendentsLayout(this);
        }
    }

    /**
     * Get current margin values from the DOM.
     */
    private static int[] collectMargin(Element elem) {
        int[] margin = { 0, 0, 0, 0 };
        margin[0] = parsePixel(getComputedStyleValue(elem, "marginTop"));
        margin[1] = parsePixel(getComputedStyleValue(elem, "marginRight"));
        margin[2] = parsePixel(getComputedStyleValue(elem, "marginBottom"));
        margin[3] = parsePixel(getComputedStyleValue(elem, "marginLeft"));
        return margin;
    }

    /**
     * Get current padding values from the DOM.
     */
    private static int[] collectPadding(Element elem) {
        int[] padding = { 0, 0, 0, 0 };
        padding[0] = parsePixel(getStyleValue(elem, "paddingTop"));
        padding[1] = parsePixel(getStyleValue(elem, "paddingRight"));
        padding[2] = parsePixel(getStyleValue(elem, "paddingBottom"));
        padding[3] = parsePixel(getStyleValue(elem, "paddingLeft"));
        return padding;
    }

    /**
     * Get current border values from the DOM.
     */
    private static int[] collectBorder(Element elem) {
        int[] border = { 0, 0, 0, 0 };
        border[0] = parsePixel(getComputedStyleValue(elem, "borderTopWidth"));
        border[1] = parsePixel(getComputedStyleValue(elem, "borderRightWidth"));
        border[2] = parsePixel(getComputedStyleValue(elem, "borderBottomWidth"));
        border[3] = parsePixel(getComputedStyleValue(elem, "borderLeftWidth"));
        return border;
    }

    /**
     * Calculates the outer size of an element, i.e. size including margins.
     * 
     * @param elem
     * @return
     */
    private static Size getOuterSize(Element elem) {
        int[] margin = collectMargin(elem);
        int width = elem.getOffsetWidth() + margin[1] + margin[3];
        int height = elem.getOffsetHeight() + margin[0] + margin[2];
        return new Size(width, height);
    }

    private static native String getComputedStyleValue(Element elem, String name)
    /*-{
        
    	// This function does not handle opacity well
    	// Name is expected in camelCase format
    	
    	// Store reference of this function (used in recursion)
    	var self = @com.vaadin.incubator.dashlayout.client.ui.VDashLayout::getComputedStyleValue(Lcom/google/gwt/user/client/Element;Ljava/lang/String;);
    	
    	// Perform normalization and checks here (e.g. if margin has value "auto", return 0)
    	if(name.indexOf("border") > -1 && name.indexOf("Width") > -1) {
    		var borderStyle = self(elem, name.substring(0,name.length-5)+"Style");
    		if(borderStyle=="none") {
    			return "0px";
    		}
    	}
    	
     	var ret, style = elem.style, force = false;
     	var defaultView = $wnd.document.defaultView || {};


    	if ( !force && style && style[ name ] ){
    	
    		// Fetch value from inline style
    		// -------------------------------------
    		ret = style[ name ];
    		
    	} else if ( defaultView.getComputedStyle ) {
    	
    		// Standards based browsers
    		// ------------------------------------

    		// Convert name to dashed format
    		name = name.replace( /([A-Z])/g, "-$1" ).toLowerCase();
    		var computedStyle = defaultView.getComputedStyle(elem, null);

    		if ( computedStyle ) {
    			ret = computedStyle.getPropertyValue(name);
    		}

    	} else if ( elem.currentStyle ) {
    	
    		// IE
    		// --------------------------------------
    		
    		var camelCase = name.replace(/\-(\w)/g, function(all, letter){
    			return letter.toUpperCase();
    		});

    		ret = elem.currentStyle[ name ];

    		// From the awesome hack by Dean Edwards
    		// http://erik.eae.net/archives/2007/07/27/18.54.15/#comment-102291

    		// If we're not dealing with a regular pixel number
    		// but a number that has a weird ending, we need to convert it to pixels
    		if ( !/^\d+(px)?$/i.test( ret ) && /^\d/.test( ret ) ) {
    			// Remember the original values
    			var left = style.left, rsLeft = elem.runtimeStyle.left;

    			// Put in the new values to get a computed value out
    			elem.runtimeStyle.left = elem.currentStyle.left;
    			style.left = ret || 0;
    			ret = style.pixelLeft + "px";

    			// Revert the changed values
    			style.left = left;
    			elem.runtimeStyle.left = rsLeft;
    		}
    	}
    	
    	// Normalize margin values
    	if(name.indexOf("margin") > -1 && ret == "auto") {
    		return "0px";
    	}
        
    	return ret;
    }-*/;

    /**
     * Normalize computed style values
     */
    private static String getStyleValue(final Element elem, final String name) {
        String ret = getComputedStyleValue(elem, name);
        if (name.equals("width") && ret.equals("auto")) {
            ret = elem.getClientWidth() + "px";
        } else if (name.equals("height") && ret.equals("auto")) {
            ret = elem.getClientHeight() + "px";
        }
        return ret;
    }

    /**
     * Takes a String value e.g. "12px" and parses that to int 12
     * 
     * @param String
     *            value with "px" ending.
     * @return int value.
     */
    private static int parsePixel(String value) {
        if (value == "" || value == null) {
            return 0;
        }
        Float ret = Float.parseFloat(value.substring(0, value.length() - 2));
        return ret.intValue();
    }

    /**
     * ChildCell represents one slot in the layout.
     */
    private class ChildCell {
        // Size is the allocated space for this widget
        private RenderSpace space = null;
        // Widget size is the natural minimum size of the widget
        private Size widgetSize = new Size(-1, -1);
        // Margins that are specified for the widget in CSS
        private int[] widgetMargin = { 0, 0, 0, 0 };
        private Widget widget;
        private FloatSize relativeSize;
        private AlignmentInfo alignment = new AlignmentInfo(1);
        private int alignmentOffsetLeft = 0;
        private int alignmentOffsetTop = 0;
        private float expandRatio = -1;
        private int surplus = 0;

        // private int top = 0;
        // private int left = 0;

        public ChildCell(Widget child) {
            widget = child;
            // Floats affect how computedStyle is calculated in browsers
            setFloat("left");
        }

        private void reset() {
            widget.setWidth("");
            widget.setHeight("");
            widget.getElement().getStyle().setPropertyPx("marginLeft",
                    widgetMargin[3]);
            widget.getElement().getStyle().setPropertyPx("marginTop",
                    widgetMargin[0]);
            alignmentOffsetLeft = 0;
            alignmentOffsetTop = 0;
            surplus = 0;
        }

        private void reAlign() {
            int left = 0;
            int top = 0;
            if (alignment != null) {
                if (alignment.isHorizontalCenter()) {
                    left = (space.getWidth() - widgetSize.getWidth()) / 2;
                } else if (alignment.isRight()) {
                    left = space.getWidth() - widgetSize.getWidth();
                }
                if (space.getHeight() > widgetSize.getHeight()) {
                    if (alignment.isVerticalCenter()) {
                        top = (space.getHeight() - widgetSize.getHeight()) / 2;
                    } else if (alignment.isBottom()) {
                        top = space.getHeight() - widgetSize.getHeight();
                    }
                }
            }
            alignmentOffsetLeft = left;
            alignmentOffsetTop = top;

            left += widgetMargin[3];
            widget.getElement().getStyle().setPropertyPx("marginLeft", left);

            top += widgetMargin[0];
            widget.getElement().getStyle().setPropertyPx("marginTop", top);

            if (BrowserInfo.get().isIE6()) {
                widget.getElement().getStyle().setProperty("display", "inline");
            }

            // Accommodate whole layout cell
            final int widgetHeight = widgetSize.getHeight() - widgetMargin[0]
                    - widgetMargin[2];
            if (widgetHeight > -1) {
                final int accommodated = widgetHeight + top + widgetMargin[2];
                if (space.getHeight() > accommodated) {
                    surplus = space.getHeight() - accommodated;
                    widget.getElement().getStyle().setPropertyPx(
                            "marginBottom", surplus);
                }
            }
        }

        private void setExpandRatio(float i) {
            expandRatio = i;
        }

        private void setAlignment(int i) {
            alignment = new AlignmentInfo(i);
        }

        private void setFloat(String floatValue) {
            widget.getElement().getStyle().setProperty("float", floatValue);
            widget.getElement().getStyle()
                    .setProperty("styleFloat", floatValue);
            widget.getElement().getStyle().setProperty("cssFloat", floatValue);
            widget.getElement().getStyle().setProperty("clear", floatValue);
            widget.getElement().getStyle().setProperty("display", "inline");
        }

        private void updateSizeInfo(UIDL uidl) {
            if (uidl.hasAttribute("cached")) {
                return;
            }
            float relativeWidth = -1;
            if (uidl.hasAttribute("width")) {
                final String w = uidl.getStringAttribute("width");
                if (w.endsWith("%")) {
                    relativeWidth = Float.parseFloat(w.substring(0,
                            w.length() - 1));
                }
            }
            float relativeHeight = -1;
            if (uidl.hasAttribute("height")) {
                final String h = uidl.getStringAttribute("height");
                if (h.endsWith("%")) {
                    relativeHeight = Float.parseFloat(h.substring(0,
                            h.length() - 1));
                }
            }
            relativeSize = new FloatSize(relativeWidth, relativeHeight);
        }

        private void updateSpace() {
            if (widget.isAttached()) {
                VDashLayout p = (VDashLayout) widget.getParent();
                final int width = parsePixel(getStyleValue(p.getElement(),
                        "width"));
                int height = -1;

                if (mode == MODE_FULL) {
                    float ratio = 0;
                    if (compoundRatio > 0 && expandRatio > 0) {
                        ratio = (expandRatio / compoundRatio);
                    } else if (compoundRatio < 0) {
                        // No expand ratios set to any component, set equal
                        // ratios to every widget
                        ratio = 1.0f / widgetToCell.size();
                    }
                    if (ratio != 0) {
                        height = p.height;
                        final int usedHeight = isRelativeHeight() ? consumedHeight
                                : consumedHeight - widgetSize.getHeight();
                        height = (int) ((height - usedHeight) * ratio);
                    } else if (isRelativeHeight()) {
                        height = 0;
                    }
                }
                space = new RenderSpace(width, height);
            }
        }

        private boolean isRelativeWidth() {
            return relativeSize.getWidth() > -1;
        }

        private boolean isRelativeHeight() {
            return relativeSize.getHeight() > -1;
        }

        private Widget getWidget() {
            return widget;
        }

        private void updateNaturalSize() {
            if (widget.isAttached()) {
                widgetSize = getOuterSize(widget.getElement());
                widgetSize
                        .setWidth(widgetSize.getWidth() - alignmentOffsetLeft);
                widgetSize.setHeight(widgetSize.getHeight()
                        - alignmentOffsetTop - surplus);
            }
        }

        private void updateWidgetMargin() {
            if (widget.isAttached()) {
                widgetMargin = collectMargin(widget.getElement());
                widgetMargin[3] -= alignmentOffsetLeft;
                widgetMargin[0] -= alignmentOffsetTop;
                widgetMargin[2] -= surplus;
            }
        }

        private RenderSpace getRenderSpace() {
            return space;
        }

        private Size getWidgetSize() {
            return widgetSize;
        }

        private RenderSpace getSpaceSansMargins() {
            return new RenderSpace(space.getWidth() - widgetMargin[1]
                    - widgetMargin[3], space.getHeight() - widgetMargin[0]
                    - widgetMargin[2]);
        }

        private boolean updateAfterOthers() {
            return isRelativeHeight() || isRelativeWidth()
                    || alignment.isBottom() || alignment.isHorizontalCenter()
                    || alignment.isRight() || alignment.isVerticalCenter()
                    || expandRatio != -1;
        }
    }

}
